/*

  rgf.c

  Author: Mika Kojo <mkojo@ssh.fi>

  Copyright (c) 1999-2000 SSH Communications Security, Finland
  All rights reserved.

  Created Thu Dec  9 21:25:47 1999.

  */

/* A library for redundancy generation functions for specific algorithms.
 */

#include "sshincludes.h"
#include "sshcrypt.h"
#include "sshcrypti.h"
#include "sshmp.h"
#include "genmp.h"
#ifdef SSHDIST_CRYPT_RSA
#ifdef WITH_RSA
#include "sshasn1.h"
#include "rsa.h"
#endif /* WITH_RSA */
#endif /* SSHDIST_CRYPT_RSA */
#ifdef SSHDIST_CRYPT_MD5
#include "md5.h"
#endif /* SSHDIST_CRYPT_MD5 */
#ifdef SSHDIST_CRYPT_SHA
#include "sha.h"
#endif /* SSHDIST_CRYPT_SHA */




#define SSH_RGF_MAXLEN  0xffffffff


/* Some useful generic hash definitions. */

SSH_RGF_HASH_ALLOCATE_FUNC(ssh_rgf_std_hash_allocate)
{
  SshRGFHash created;

  if (def == NULL)
    ssh_fatal("ssh_rgf_std_hash_allocate: no hash definition.");
  created           = ssh_xcalloc(1, sizeof(*created));
  created->def      = def;
  created->hash_def = hash_def;

  if (hash_def != NULL)
    {
      created->context = ssh_hash_allocate_internal(hash_def);
      if (created->context == NULL)
        {
          ssh_xfree(created);
          return NULL;
        }
      ssh_hash_reset(created->context);
    }
  else
    {
      ssh_xfree(created);
      return NULL;
    }
  return created;
}

SSH_RGF_HASH_FREE_FUNC(ssh_rgf_std_hash_free)
{
  ssh_hash_free(hash->context);
  ssh_xfree(hash);
}

SSH_RGF_HASH_UPDATE_FUNC(ssh_rgf_std_hash_update)
{
  if (hash->context == NULL)
    ssh_fatal("ssh_rgf_std_hash_update: no hash state.");

  /* Handle the case when possibly setting the finalized digest
     beforehand. */
  if (for_digest)
    {
      if (ssh_hash_digest_length(hash->context) == data_len)
        {
          /* Twhis does not allocate new space for the data. */
          hash->precomp_digest        = data;
          hash->precomp_digest_length = data_len;
          return TRUE;
        }
      return FALSE;
    }

  if (hash->precomp_digest)
    return FALSE;

  ssh_hash_update(hash->context, data, data_len);
  return TRUE;
}

SSH_RGF_HASH_DIGEST_LENGTH_FUNC(ssh_rgf_std_hash_digest_length)
{
  if (hash->context == NULL)
    ssh_fatal("ssh_rgf_std_hash_digest_length: no hash state.");
  return ssh_hash_digest_length(hash->context);
}

SSH_RGF_HASH_FINALIZE_FUNC(ssh_rgf_std_hash_finalize)
{
  if (hash->context == NULL)
    ssh_fatal("ssh_rgf_std_hash_finalize: no hash state.");
  if (hash->precomp_digest)
    memcpy(digest, hash->precomp_digest, hash->precomp_digest_length);
  else
    ssh_hash_final(hash->context, digest);
}

SSH_RGF_HASH_ASN1_OID_FUNC(ssh_rgf_std_hash_asn1_oid)
{
  if (hash->context == NULL)
    ssh_fatal("ssh_rgf_std_hash_asn1_oid: no hash state.");
  return ssh_hash_asn1_oid(hash->context);
}

/* With no hashing. */

SSH_RGF_HASH_ALLOCATE_FUNC(ssh_rgf_none_hash_allocate)
{
  SshRGFHash created;

  if (def == NULL)
    ssh_fatal("ssh_rgf_none_hash_allocate: no hash definition.");

  created = ssh_xcalloc(1, sizeof(*created));
  created->def      = def;
  created->hash_def = NULL;
  return created;
}

SSH_RGF_HASH_FREE_FUNC(ssh_rgf_none_hash_free)
{
  ssh_xfree(hash);
}

SSH_RGF_HASH_UPDATE_FUNC(ssh_rgf_none_hash_update)
{
  hash->precomp_digest        = data;
  hash->precomp_digest_length = data_len;
  return TRUE;
}

SSH_RGF_HASH_DIGEST_LENGTH_FUNC(ssh_rgf_none_hash_digest_length)
{
  if (hash->precomp_digest == NULL)
    return SSH_RGF_MAXLEN;
  return hash->precomp_digest_length;
}

SSH_RGF_HASH_FINALIZE_FUNC(ssh_rgf_none_hash_finalize)
{
  if (hash->precomp_digest)
    memcpy(digest, hash->precomp_digest, hash->precomp_digest_length);
}

SSH_RGF_HASH_ASN1_OID_FUNC(ssh_rgf_none_hash_asn1_oid)
{
  return NULL;
}

/* Basic hash definitions. */
const SshRGFHashDef ssh_rgf_std_hash_def =
{
  ssh_rgf_std_hash_free,
  ssh_rgf_std_hash_update,
  ssh_rgf_std_hash_digest_length,
  ssh_rgf_std_hash_finalize,
  ssh_rgf_std_hash_asn1_oid,
  NULL,
  NULL
};

const SshRGFHashDef ssh_rgf_none_hash_def =
{
  ssh_rgf_none_hash_free,
  ssh_rgf_none_hash_update,
  ssh_rgf_none_hash_digest_length,
  ssh_rgf_none_hash_finalize,
  ssh_rgf_none_hash_asn1_oid,
  NULL,
  NULL
};

/* Basic redundancy function implementations. */

#ifdef SSHDIST_CRYPT_RSA
#ifdef WITH_RSA
/* RSA PKCS-1 v1.5 */

/* Some useful routines doing the dirty work. */
SSH_RGF_ENCRYPT_FUNC(ssh_rgf_pkcs1_encrypt)
{
  SshMPIntStruct t1, t2;
  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Remark. This function could be written to run without the
     integer arithmetic, but the padding routines use them so... */

  /* Convert the message into an integer. */
  ssh_buf_to_mp(&t1, msg, msg_len);

  /* Now handle the padding.
   */
  ssh_rsa_pkcs1_pad_data(&t2, &t1, (unsigned int)msg_len, 2,
                         output_msg_len);

  /* Convert back to an plaintext buffer. */
  ssh_mp_to_buf(output_msg, output_msg_len, &t2);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return SSH_RGF_OK;
}

SSH_RGF_DECRYPT_FUNC(ssh_rgf_pkcs1_decrypt)
{
  SshMPIntStruct t1;
  unsigned char *buf;
  size_t         buf_len;

  ssh_mp_init(&t1);
  ssh_buf_to_mp(&t1, decrypted_msg, decrypted_msg_len);

  /* Allocate a temporary buffer. */
  buf = ssh_xmalloc(max_output_msg_len);

  /* Unpad please. */
  if (ssh_rsa_pkcs1_unpad_data(2, &t1,
                               buf, max_output_msg_len,
                               &buf_len) == FALSE)
    {
      ssh_xfree(buf);
      ssh_mp_clear(&t1);
      return SSH_RGF_OP_FAILED;
    }

  /* Return the unpadded msg. */
  *output_msg     = buf;
  *output_msg_len = buf_len;

  ssh_buf_to_mp(&t1, decrypted_msg, decrypted_msg_len);
  ssh_mp_clear(&t1);

  return SSH_RGF_OK;
}

SSH_RGF_SIGN_FUNC(ssh_rgf_pkcs1_sign)
{
  SshAsn1Context asn1_context;
  SshAsn1Status  status;
  SshAsn1Tree    tree;
  unsigned char *ber_data;
  size_t         ber_data_len;
  unsigned char *digest;
  const char    *oid;
  size_t         digest_len;
  SshMPIntStruct         t1, t2;

  if (hash->context == NULL)
    return SSH_RGF_OP_FAILED;

  /* Get the hash or msg context. */

  /* Finalize the hash function. */
  digest_len = (*hash->def->rgf_hash_digest_length)(hash);
  digest     = ssh_xmalloc(digest_len);
  oid        = (*hash->def->rgf_hash_asn1_oid)(hash);
  (*hash->def->rgf_hash_finalize)(hash, digest);

  if (oid == NULL)
    return SSH_RGF_OP_FAILED;

  /* Make a ASN.1/BER encoded string. */
  asn1_context = ssh_asn1_init();


  /* Build simple ASN.1 tree which contains just basic sequences and
     object identifiers and octet strings. If speed is mandatory we
     could precompute parts of the der encoding and place appropriate
     lengths here. */
  status = ssh_asn1_create_tree(asn1_context, &tree,
                                "(sequence ()"
                                "  (sequence ()"
                                "    (object-identifier ())"
                                "    (null ()))"
                                "  (octet-string ()))",
                                oid,
                                digest, digest_len);

  /* Free the allocated digest. */
  ssh_xfree(digest);

  /* Verify that the tree was correctly reconstructed. */
  if (status != SSH_ASN1_STATUS_OK)
    {
      ssh_asn1_free(asn1_context);
      return SSH_RGF_OP_FAILED;
    }

  /* Do the encoding. */
  status = ssh_asn1_encode(asn1_context, tree);
  if (status != SSH_ASN1_STATUS_OK)
    {
      ssh_asn1_free(asn1_context);
      return SSH_RGF_OP_FAILED;
    }

  /* Get BER encoded blob. */
  ssh_asn1_get_data(tree, &ber_data, &ber_data_len);
  /* Free ASN.1 context. */
  ssh_asn1_free(asn1_context);

  /* Compute signature. */

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, ber_data, ber_data_len);

  /* Free the ber data allocate by ssh_asn1_get_data. */
  ssh_xfree(ber_data);

  /* Pad data as pkcs1 suggests. */
  if (output_msg_len < ber_data_len + 1)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return SSH_RGF_OP_FAILED;
    }

  /* Use the PKCS-1 padding. */
  ssh_rsa_pkcs1_pad_data(&t2, &t1, (unsigned int)ber_data_len,
                         1, output_msg_len);

  /* Convert back to octet string. */
  ssh_mp_to_buf(output_msg, output_msg_len, &t2);

  /* Free the hash or msg data. */
  ssh_rgf_hash_free(hash);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return SSH_RGF_OK;
}

SSH_RGF_VERIFY_FUNC(ssh_rgf_pkcs1_verify)
{
  SshAsn1Context asn1_context;
  SshAsn1Status status;
  SshAsn1Tree tree;
  unsigned char *ber_buf, *temp_buf;
  SshMPIntStruct t1;
  char *oid;
  const char *hash_oid;
  unsigned char *digest;
  size_t digest_len, return_len;
  Boolean rv;

  if (hash->context == NULL)
    return SSH_RGF_OP_FAILED;

  *output_msg     = NULL;
  *output_msg_len = 0;

  /* Be a pessimist. */
  rv = FALSE;

  ssh_mp_init(&t1);

  /* Decode the msg. */
  ber_buf = ssh_xmalloc(max_output_msg_len);

  ssh_buf_to_mp(&t1, decrypted_signature, decrypted_signature_len);

  /* Unpad. */
  if (ssh_rsa_pkcs1_unpad_data(1, &t1,
                               ber_buf, max_output_msg_len,
                               &return_len) == FALSE)
    {
      ssh_xfree(ber_buf);
      ssh_mp_clear(&t1);
      goto failed;
    }

  /* Clear the multiple precision integer. */
  ssh_mp_clear(&t1);

  /* Now we have the "pleasant" job of decoding ASN.1 BER coded blob. */
  asn1_context = ssh_asn1_init();

  status = ssh_asn1_decode(asn1_context, ber_buf, return_len, &tree);

  /* Free BER buffer. */
  ssh_xfree(ber_buf);

  if (status != SSH_ASN1_STATUS_OK)
    {
      ssh_asn1_free(asn1_context);
      goto failed;
    }

  status = ssh_asn1_read_tree(asn1_context, tree,
                              "(sequence ()"
                              "  (sequence ()"
                              "    (object-identifier ())"
                              "    (null ()))"
                              "  (octet-string ()))",
                              &oid,
                              &temp_buf, &return_len);

  /* Free asn1 context. */
  ssh_asn1_free(asn1_context);
  if (status != SSH_ASN1_STATUS_OK)
    goto failed;

  /* Hash if needed. */
  digest_len = (*hash->def->rgf_hash_digest_length)(hash);
  digest = ssh_xmalloc(digest_len);
  (*hash->def->rgf_hash_finalize)(hash, digest);
  hash_oid = (*hash->def->rgf_hash_asn1_oid)(hash);

  /* Compare. */
  if (hash_oid == NULL || digest_len != return_len)
    {
      ssh_xfree(oid);
      ssh_xfree(temp_buf);
      ssh_xfree(digest);
      goto failed;
    }
  if (memcmp(digest, temp_buf, digest_len) == 0 &&
      strcmp(hash_oid, oid) == 0)
    rv = TRUE;

  ssh_xfree(oid);
  ssh_xfree(digest);
  ssh_xfree(temp_buf);
failed:

  ssh_rgf_hash_free(hash);

  if (rv == TRUE)
    return SSH_RGF_OK;
  return SSH_RGF_OP_FAILED;
}

SSH_RGF_SIGN_FUNC(ssh_rgf_pkcs1_sign_nohash)
{
  SshMPIntStruct t1, t2;
  unsigned char *digest;
  size_t digest_length;

  digest_length = (*hash->def->rgf_hash_digest_length)(hash);
  if (digest_length == SSH_RGF_MAXLEN)
    return SSH_RGF_OP_FAILED;

  digest        = ssh_xmalloc(digest_length);
  (*hash->def->rgf_hash_finalize)(hash, digest);

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Convert to integer. */
  ssh_buf_to_mp(&t1, digest, digest_length);

  /* Pad data. */
  ssh_rsa_pkcs1_pad_data(&t2, &t1, (unsigned int)digest_length,
                         1, output_msg_len);

  ssh_mp_to_buf(output_msg, output_msg_len, &t2);

  /* Free the "hash". */
  ssh_rgf_hash_free(hash);
  ssh_xfree(digest);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return SSH_RGF_OK;
}

SSH_RGF_VERIFY_FUNC(ssh_rgf_pkcs1_verify_nohash)
{
  SshMPIntStruct t1;
  unsigned char *buf;
  size_t return_len;
  unsigned char *digest;
  size_t digest_length;

  *output_msg     = NULL;
  *output_msg_len = 0;

  ssh_mp_init(&t1);

  ssh_buf_to_mp(&t1, decrypted_signature, decrypted_signature_len);

  /* Allocate a suitable decoding buffer. */
  buf = ssh_xmalloc(max_output_msg_len);

  /* Unpad. */
  if (ssh_rsa_pkcs1_unpad_data(1, &t1,
                               buf, max_output_msg_len,
                               &return_len) == FALSE)
    {
      ssh_xfree(buf);
      ssh_mp_clear(&t1);
      return SSH_RGF_OP_FAILED;
    }

  ssh_mp_clear(&t1);

  digest_length = (*hash->def->rgf_hash_digest_length)(hash);
  digest        = ssh_xmalloc(digest_length);
  (*hash->def->rgf_hash_finalize)(hash, digest);

  if (digest_length != return_len ||
      memcmp(digest, buf, digest_length) != 0)
    {
      ssh_xfree(buf);
      ssh_xfree(digest);
      return SSH_RGF_OP_FAILED;
    }

  ssh_xfree(buf);
  ssh_xfree(digest);
  ssh_rgf_hash_free(hash);

  return SSH_RGF_OK;
}

/* RSA PKCS-1 v2.0 */

SSH_RGF_ENCRYPT_FUNC(ssh_rgf_pkcs1v2_encrypt)
{
  unsigned char *param;
  size_t param_len;

  if (rgf->hash_def == NULL)
    return SSH_RGF_OP_FAILED;

  param = ssh_rsa_pkcs1v2_default_explicit_param(rgf->hash_def, &param_len);
  if (param == NULL)
    return SSH_RGF_OP_FAILED;

  if (output_msg_len == 0)
    return SSH_RGF_OP_FAILED;
  /* Initialize the highest octet. */
  output_msg[0] = 0;

  if (ssh_rsa_oaep_encode_with_mgf1(rgf->hash_def,
                                    msg, msg_len,
                                    param, param_len,
                                    output_msg+1, output_msg_len-1) == FALSE)
    {
      ssh_xfree(param);
      return SSH_RGF_OP_FAILED;
    }
  ssh_xfree(param);

  return SSH_RGF_OK;
}

SSH_RGF_DECRYPT_FUNC(ssh_rgf_pkcs1v2_decrypt)
{
  unsigned char *param;
  size_t param_len;

  if (rgf->hash_def == NULL)
    return SSH_RGF_OP_FAILED;

  /* Find params. */
  param = ssh_rsa_pkcs1v2_default_explicit_param(rgf->hash_def, &param_len);
  if (param == NULL)
    return SSH_RGF_OP_FAILED;

  if (decrypted_msg_len == 0 ||
      decrypted_msg[0] != 0)
    return SSH_RGF_OP_FAILED;

  /* Apply the OAEP decoding. */
  if (ssh_rsa_oaep_decode_with_mgf1(rgf->hash_def,
                                    decrypted_msg+1, decrypted_msg_len-1,
                                    param, param_len,
                                    output_msg, output_msg_len) == FALSE)
    {
      ssh_xfree(param);
      return SSH_RGF_OP_FAILED;
    }
  ssh_xfree(param);
  return SSH_RGF_OK;
}

#ifdef SSHDIST_CRYPT_SHA
/* RSA PKCS-1 v1.5 */

const SshRGFDef ssh_rgf_pkcs1_sha1_def =
{
  "pkcs1-sha1",
  ssh_rgf_pkcs1_encrypt,
  ssh_rgf_pkcs1_decrypt,
  ssh_rgf_pkcs1_sign,
  ssh_rgf_pkcs1_verify,
  NULL,
  ssh_rgf_std_hash_allocate,
  &ssh_rgf_std_hash_def,
  &ssh_hash_sha_def,
  NULL
};
#endif /* SSHDIST_CRYPT_SHA */

#ifdef SSHDIST_CRYPT_MD5

const SshRGFDef ssh_rgf_pkcs1_md5_def =
{
  "pkcs1-md5",
  ssh_rgf_pkcs1_encrypt,
  ssh_rgf_pkcs1_decrypt,
  ssh_rgf_pkcs1_sign,
  ssh_rgf_pkcs1_verify,
  NULL,
  ssh_rgf_std_hash_allocate,
  &ssh_rgf_std_hash_def,
  &ssh_hash_md5_def,
  NULL
};

#endif /* SSHDIST_CRYPT_MD5 */
















const SshRGFDef ssh_rgf_pkcs1_none_def =
{
  "pkcs1-none",
  ssh_rgf_pkcs1_encrypt,
  ssh_rgf_pkcs1_decrypt,
  ssh_rgf_pkcs1_sign_nohash,
  ssh_rgf_pkcs1_verify_nohash,
  NULL,
  ssh_rgf_none_hash_allocate,
  &ssh_rgf_none_hash_def,
  NULL,
  NULL
};

#ifdef SSHDIST_CRYPT_SHA

/* RSA PKCS-1 v2.0 */

const SshRGFDef ssh_rgf_pkcs1v2_sha1_def =
{
  "pkcs1v2-sha1",
  ssh_rgf_pkcs1v2_encrypt,
  ssh_rgf_pkcs1v2_decrypt,
  ssh_rgf_pkcs1_sign,
  ssh_rgf_pkcs1_verify,
  NULL,
  ssh_rgf_std_hash_allocate,
  &ssh_rgf_std_hash_def,
  &ssh_hash_sha_def,
  NULL
};
#endif /* SSHDIST_CRYPT_SHA */

#ifdef SSHDIST_CRYPT_MD5

const SshRGFDef ssh_rgf_pkcs1v2_md5_def =
{
  "pkcs1v2-md5",
  ssh_rgf_pkcs1v2_encrypt,
  ssh_rgf_pkcs1v2_decrypt,
  ssh_rgf_pkcs1_sign,
  ssh_rgf_pkcs1_verify,
  NULL,
  ssh_rgf_std_hash_allocate,
  &ssh_rgf_std_hash_def,
  &ssh_hash_md5_def,
  NULL
};
#endif /* SSHDIST_CRYPT_MD5 */


















const SshRGFDef ssh_rgf_pkcs1v2_none_def =
{
  "pkcs1v2-none",
  ssh_rgf_pkcs1v2_encrypt,
  ssh_rgf_pkcs1v2_decrypt,
  ssh_rgf_pkcs1_sign_nohash,
  ssh_rgf_pkcs1_verify_nohash,
  NULL,
  ssh_rgf_none_hash_allocate,
  &ssh_rgf_none_hash_def,
  NULL,
  NULL
};
#endif /* WITH_RSA */
#endif /* SSHDIST_CRYPT_RSA */

/* A generic routines that can be used with many cryptosystems with
   little redundancy management. These include e.g. the DSA algorithm.

   Common idea with all the methods is that they basically do not
   do any redundancy related operations. For example, they just hash
   the message for signature using one of the standard hash functions.
   They do not pad the digest before signing, usually because these
   methods include the digest into the cryptosystem in more complicated
   manner than RSA does, for example.
   */

SSH_RGF_ENCRYPT_FUNC(ssh_rgf_std_encrypt)
{
  if (msg_len > output_msg_len)
    return SSH_RGF_OP_FAILED;

  /* Zero the output msg. */
  memset(output_msg, 0, output_msg_len);
  memcpy(output_msg + (output_msg_len - msg_len), msg, msg_len);

  return SSH_RGF_OK;
}

SSH_RGF_DECRYPT_FUNC(ssh_rgf_std_decrypt)
{
  if (decrypted_msg_len > max_output_msg_len)
    return SSH_RGF_OP_FAILED;
  *output_msg = ssh_xmemdup(decrypted_msg, decrypted_msg_len);
  *output_msg_len = decrypted_msg_len;
  return SSH_RGF_OK;
}

SSH_RGF_SIGN_FUNC(ssh_rgf_std_sign)
{
  unsigned char *digest;
  size_t         digest_len;

  if (hash->context == NULL)
    return SSH_RGF_OP_FAILED;

  /* Create the digest. */
  digest_len = (*hash->def->rgf_hash_digest_length)(hash);
  if (digest_len == SSH_RGF_MAXLEN)
    return SSH_RGF_OP_FAILED;
  digest     = ssh_xmalloc(digest_len);
  (*hash->def->rgf_hash_finalize)(hash, digest);

  /* Now check whether we can output the digest or not. */
  if (digest_len > output_msg_len)
    {
      ssh_xfree(digest);
      return SSH_RGF_OP_FAILED;
    }
  memset(output_msg, 0, output_msg_len);
  memcpy(output_msg + (output_msg_len - digest_len), digest, digest_len);
  ssh_xfree(digest);

  ssh_rgf_hash_free(hash);

  return SSH_RGF_OK;
}

SSH_RGF_SIGN_FUNC(ssh_rgf_std_sign_no_hash)
{
  const unsigned char *msg;
  size_t msg_len;
  unsigned char *digest;
  size_t digest_len;

  /* Create the digest. */
  digest_len = (*hash->def->rgf_hash_digest_length)(hash);
  if (digest_len == SSH_RGF_MAXLEN)
    return SSH_RGF_OP_FAILED;
  digest     = ssh_xmalloc(digest_len);
  (*hash->def->rgf_hash_finalize)(hash, digest);

  msg     = digest;
  msg_len = digest_len;

  if (msg_len > output_msg_len)
    {
      msg += (msg_len - output_msg_len);
      msg_len = output_msg_len;
    }

  memset(output_msg, 0, output_msg_len);
  memcpy(output_msg + (output_msg_len - msg_len),
         msg, msg_len);

  ssh_xfree(digest);
  ssh_rgf_hash_free(hash);
  return SSH_RGF_OK;
}

SSH_RGF_VERIFY_FUNC(ssh_rgf_std_verify)
{
  unsigned char *digest;
  size_t         digest_len;

  *output_msg     = NULL;
  *output_msg_len = 0;

  if (hash->context == NULL)
    return SSH_RGF_OP_FAILED;

  /* Create the digest. */
  digest_len = (*hash->def->rgf_hash_digest_length)(hash);
  if (digest_len == SSH_RGF_MAXLEN)
    return SSH_RGF_OP_FAILED;
  digest     = ssh_xmalloc(digest_len);
  (*hash->def->rgf_hash_finalize)(hash, digest);

  if (digest_len != decrypted_signature_len ||
      memcmp(decrypted_signature, digest, digest_len) != 0)
    {
      ssh_xfree(digest);
      return SSH_RGF_OP_FAILED;
    }
  ssh_xfree(digest);
  ssh_rgf_hash_free(hash);
  return SSH_RGF_OK;
}

SSH_RGF_VERIFY_FUNC(ssh_rgf_std_verify_no_hash)
{
  SshRGFStatus rv = SSH_RGF_OK;
  unsigned char *digest;
  size_t digest_len;

  *output_msg     = NULL;
  *output_msg_len = 0;

  /* Create the digest. */
  digest_len = (*hash->def->rgf_hash_digest_length)(hash);
  if (digest_len == SSH_RGF_MAXLEN)
    return SSH_RGF_OP_FAILED;
  digest     = ssh_xmalloc(digest_len);
  (*hash->def->rgf_hash_finalize)(hash, digest);

  /* Check the validity. */
  if (digest_len != decrypted_signature_len ||
      memcmp(digest, decrypted_signature, digest_len) != 0)
    rv = SSH_RGF_OP_FAILED;

  ssh_xfree(digest);
  ssh_rgf_hash_free(hash);
  return rv;
}

#ifdef SSHDIST_CRYPT_SHA

const SshRGFDef ssh_rgf_std_sha1_def =
{
  "std-sha1",
  ssh_rgf_std_encrypt,
  ssh_rgf_std_decrypt,
  ssh_rgf_std_sign,
  ssh_rgf_std_verify,
  NULL,
  ssh_rgf_std_hash_allocate,
  &ssh_rgf_std_hash_def,
  &ssh_hash_sha_def,
  NULL
};
#endif /* SSHDIST_CRYPT_SHA */

#ifdef SSHDIST_CRYPT_MD5

const SshRGFDef ssh_rgf_std_md5_def =
{
  "std-md5",
  ssh_rgf_std_encrypt,
  ssh_rgf_std_decrypt,
  ssh_rgf_std_sign,
  ssh_rgf_std_verify,
  NULL,
  ssh_rgf_std_hash_allocate,
  &ssh_rgf_std_hash_def,
  &ssh_hash_md5_def,
  NULL
};

#endif /* SSHDIST_CRYPT_MD5 */


















const SshRGFDef ssh_rgf_dummy_def =
{
  "std-dummy",
  ssh_rgf_std_encrypt,
  ssh_rgf_std_decrypt,
  ssh_rgf_std_sign_no_hash,
  ssh_rgf_std_verify_no_hash,
  NULL,
  ssh_rgf_none_hash_allocate,
  &ssh_rgf_none_hash_def,
  NULL,
  NULL
};

/* TODO. More redundancy functions and similar constructs. */

/* The RGF manipulation functions. */

SshRGFHash
ssh_rgf_hash_allocate_internal(const SshRGFDef *rgf_def)
{
  if (rgf_def->rgf_hash_allocate == NULL ||
      rgf_def->rgf_hash_def == NULL)
    return NULL;

  return (*rgf_def->rgf_hash_allocate)(rgf_def->rgf_hash_def,
                                       rgf_def->hash_def,
                                       rgf_def->context);
}

void
ssh_rgf_hash_update(SshRGFHash hash,
                    const unsigned char *data, size_t data_len)
{
  if (hash->def->rgf_hash_update == NULL)
    ssh_fatal("ssh_rgf_hash_update: no update function defined.");
  (*hash->def->rgf_hash_update)(hash, FALSE,
                                data, data_len);
}

Boolean
ssh_rgf_hash_update_with_digest(SshRGFHash hash,
                                const unsigned char *data, size_t data_len)
{
  if (hash->def->rgf_hash_update == NULL)
    ssh_fatal("ssh_rgf_hash_update_with_digest: no update function defined.");
  return (*hash->def->rgf_hash_update)(hash, TRUE,
                                       data, data_len);
}

SshHash
ssh_rgf_hash_derive(SshRGFHash hash)
{
  /* Check whether the conversion is possible. */
  if (hash->hash_def == NULL)
    return NULL;
  return ssh_hash_allocate_internal(hash->hash_def);
}

void
ssh_rgf_hash_free(SshRGFHash hash)
{
  if (hash->def->rgf_hash_free == NULL)
    ssh_fatal("ssh_rgf_hash_free: no hash free function defined.");
  (*hash->def->rgf_hash_free)(hash);
}

/* rgf.c */
